# cython: language_level=3
cdef class CPathfinderError(Exception):
    pass

cdef class CDublicateError(CPathfinderError):
    pass

cdef class CPathError(CPathfinderError):
    pass

cdef class LowComby:
    cdef list list_
    cdef set set_
    cpdef LowComby copy(self)    
    cpdef append(self, CNode obj)   
    cpdef insert(self, int index, CNode obj)   
    cpdef pop(self, int index)

cdef class CNode:
    cdef public double cost, probable_cost
    cdef readonly dict _connections
    cdef public object id    
    cpdef void connect(self, dict conn, bint reflect = *)
    cpdef CNode _copy(self, dict nodes)

cdef list construct(CNode startCNode, CNode endCNode)
cdef wrappop(LowComby x)
cdef wrappush(LowComby x,CNode n)
cpdef list djikstra_bestpath(CNode startCNode, CNode endCNode, bint first_contact = *)
cpdef list astar_bestpath(CNode startCNode, CNode endCNode, object func,  bint first_contact = *, list args = *)
cpdef tuple copy_graph(CNode start, CNode stop)


from heapq import heappop, heappush
from typing import Callable, Dict, Hashable, List, Union
from pypathfinder.utils import LowComby, HighComby, get_pop, get_push
from pypathfinder.Dijkstra import Node, construct

class ANode(Node):
    def __init__(self, id: Hashable, connections: Dict["ANode", int] = None):
        super().__init__(id, connections)
        self.probable_cost = float("inf")
        self._connections: Dict[ANode, int]
    
    def __str__(self) -> str:
        return f"Node(id={self.id}, cost={self.cost}, probable_cost={self.probable_cost})"
    
    def __hash__(self) -> int:
        return super().__hash__()
    
    def __lt__(self, other):
        if not isinstance(other, type(self)):
            raise NotImplemented
        return self.probable_cost < other.probable_cost
    
    def __le__(self, other):
        if not isinstance(other, type(self)):
            raise NotImplemented
        return self.probable_cost <= other.probable_cost
    
    def __gt__(self, other):
        if not isinstance(other, type(self)):
            raise NotImplemented
        return self.probable_cost > other.probable_cost

    def __ge__(self, other):
        if not isinstance(other, type(self)):
            raise NotImplemented
        return self.probable_cost >= other.probable_cost
    
    def connect(self, conn: Dict["ANode", int], reflect: bool = False):
        return super().connect(conn, reflect)
    

def bestpath(startnode: ANode, endnode: ANode, func: Callable, first_contact: bool = False, queue_type: Union[list, LowComby, HighComby]=list, args: list = None) -> List[ANode]:
    """
    Args:
        startnode: starting point
        endnode: ending point
        first_contact: if True, pathfinding will end as soon as the endnode has first been discoverd
        queue_type: type used for queueing nodes; LowComby can be faster but consumes more memory
    
    Returns:
        list of nodes, creating a path
    """
    if args is None:
        args = []
    
    # get costs
    startnode.cost = startnode.probable_cost = 0
    queue: Union[List[ANode], LowComby] = queue_type()
    queue.append(startnode)
    use_heappop = get_pop(queue)
    use_heappush = get_push(queue)

    while len(queue) != 0:
        currentnode = use_heappop(queue)
        for node, cost in currentnode._connections.items():
                
            new_cost = currentnode.cost + cost
            if new_cost < node.cost:
                node.cost = new_cost
                node.probable_cost = new_cost+ func(node, args)
                if node not in queue:
                    use_heappush(queue, node)
                    #queue.append(node)

        if endnode.cost <= currentnode.probable_cost:
            break
        if first_contact and currentnode is endnode:
            break
    
    # get path
    return construct(startnode, endnode)



from heapq import heappop, heappush


class PathfinderError(Exception): pass

class DublicateError(PathfinderError): pass

class PathError(PathfinderError): pass

class LowComby:
    def __init__(self) -> None:
        self.list = []
        self.set = set()
    
    def __getitem__(self, index):
        return self.list[index]
    
    def __setitem__(self, index: int, obj):
        self.list[index] = obj
        self.set.add(obj)
    
    def __len__(self):
        return len(self.list)
    
    def __contains__(self, obj):
        return obj in self.set
    
    def copy(self):
        lc = LowComby()
        lc.list = self.list.copy()
        lc.set = self.set.copy()
        return lc
    
    def append(self, obj):
        self.list.append(obj)
        self.set.add(obj)
    
    def insert(self, index: int, obj):
        self.list.insert(index, obj)
        self.set.add(obj)
    
    def pop(self, index):
        r = self.list.pop(index)
        self.set.pop(r)
        return r
    
    def sort(self, key=None):
        if key is None:
            key = lambda x: x.cost
        self.list.sort(key=key)

class HighComby(LowComby):
    def __getitem__(self, index):
        if index < 0 or index >= len(self.list):
            raise ValueError("index out of bounds")
        return super().__getitem__(index)
    
    def __setitem__(self, index: int, obj):
        if obj in self.set and self.list[index] != obj:
            raise DublicateError("object already exists")
        if index < 0 or index >= len(self.list):
            raise IndexError("index out of bounds")
        if not hasattr(obj, "__hash__"):
            raise TypeError("object can't be hashed")
        super().__setitem__(index, obj)
    
    def append(self, obj):
        if obj in self.set:
            raise DublicateError("object already exists")
        if not hasattr(obj, "__hash__"):
            raise TypeError("object can't be hashed")
        super().append(obj)
    
    def insert(self, index: int, obj):
        if obj in self.set and self.list[index] != obj:
            raise DublicateError("object already exists")
        if index < 0 or index >= len(self.list):
            raise IndexError("index out of bounds")
        if not hasattr(obj, "__hash__"):
            raise TypeError("object can't be hashed")
        
        super().insert(index, obj)
    
    def pop(self, index):
        try:
            r = self.list.pop(index)
            self.set.pop(r)
            return r
        except IndexError as e:
            raise IndexError("out of bounds") from e
        except KeyError as e:
            raise PathfinderError("set didn't contain item") from e

def get_pop(queue):
    return heappop if isinstance(queue, list) else lambda x: heappop(x.list)

def get_push(queue):
    return heappush if isinstance(queue, list) else lambda x,n: heappush(x.list, n)



# cython: language_level=3, emit_code_comments=True, embedsignature=True
from heapq import heappop, heappush
cimport cython

cdef class CPathfinderError(Exception): pass

cdef class CDublicateError(CPathfinderError): pass

cdef class CPathError(CPathfinderError): pass

cdef double INFINITY = float("inf")

cdef class LowComby:
    def __init__(self):
        self.list_ = []
        self.set_ = set()
    
    def __getitem__(self, int index):
        return self.list[index]
    
    def __setitem__(self, int index, CNode obj):
        self.list_[index] = obj
        self.set_.add(obj)
    
    def __len__(self):
        return len(self.list_)
    
    def __contains__(self, CNode obj):
        return obj in self.set_
    
    cpdef LowComby copy(self):
        cdef LowComby lc
        lc = LowComby()
        lc.list_ = self.list_.copy()
        lc.set = self.set_.copy()
        return lc
    
    cpdef append(self, CNode obj):
        self.list_.append(obj)
        self.set_.add(obj)
    
    cpdef insert(self, int index, CNode obj):
        self.list_.insert(index, obj)
        self.set_.add(obj)
    
    cpdef pop(self, int index):
        r = self.list_.pop(index)
        self.set_.remove(r)
        return r

class HighComby(LowComby):
    def __getitem__(self, index):
        if index < 0 or index >= len(self.list):
            raise ValueError("index out of bounds")
        return super().__getitem__(index)
    
    def __setitem__(self, index: int, obj):
        if obj in self.set and self.list[index] != obj:
            raise CDublicateError("object already exists")
        if index < 0 or index >= len(self.list):
            raise IndexError("index out of bounds")
        if not hasattr(obj, "__hash__"):
            raise TypeError("object can't be hashed")
        super().__setitem__(index, obj)
    
    def append(self, obj):
        if obj in self.set:
            raise CDublicateError("object already exists")
        if not hasattr(obj, "__hash__"):
            raise TypeError("object can't be hashed")
        super().append(obj)
    
    def insert(self, int index, obj):
        if obj in self.set and self.list[index] != obj:
            raise CDublicateError("object already exists")
        if index < 0 or index >= len(self.list):
            raise IndexError("index out of bounds")
        if not hasattr(obj, "__hash__"):
            raise TypeError("object can't be hashed")
        
        super().insert(index, obj)
    
    def pop(self, int index):
        try:
            r = self.list.pop(index)
            self.set.pop(r)
            return r
        except IndexError as e:
            raise IndexError("out of bounds") from e
        except KeyError as e:
            raise CPathfinderError("set didn't contain item") from e

cdef wrappop(LowComby x):
    return heappop(x.list_)

cdef wrappush(LowComby x, CNode n):
    return heappush(x.list_, n)

cdef class CNode:
    def __init__(self, id_, connections = None):
        self._connections = {} if connections is None else connections
        self.id = id_
        self.cost = INFINITY
        self.probable_cost = INFINITY
    
    def __str__(self) -> str:
        return f"CNode(id={self.id}, cost={self.cost})"
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def __lt__(self, CNode other):
        return self.cost < other.cost
    
    def __le__(self, CNode other):
        return self.cost <= other.cost
    
    def __gt__(self, CNode other):
        return self.cost > other.cost

    def __ge__(self, CNode other):
        return self.probable_cost >= other.probable_cost
    
    def __eq__(self, CNode other):
        return other.id == self.id

    def __hash__(self) -> int:
        return hash(self.id)
    
    cpdef void connect(self, dict conn, bint reflect = False):
        """
        Args:
            conn: connections
            reflect: if the connections should also be applied to the connected CNodes
        """
        cdef double value
        cdef CNode cnode
        self._connections.update(conn)
        if reflect:
            for cnode, value in conn.items():
                cnode.connect({self: value}, False)
    
    cpdef CNode _copy(self, dict nodes): 
        cdef CNode self_copy = CNode(self.id)
        cdef dict new_connections = {}
        cdef CNode node, node_copy
        cdef double cost
        nodes[self_copy] = self_copy

        for node, cost in self._connections.items():
            if node in nodes.keys():
                node_copy = nodes.get(node)
            else:
                node_copy = node._copy(nodes)
                nodes[node_copy] = node_copy
            new_connections[node_copy] = cost
        self_copy.connect(new_connections, False)
        return self_copy

cdef list construct(CNode startCNode, CNode endCNode):
    # get path
    cdef list path
    cdef CNode to_check
    cdef CNode cnode

    path = [endCNode]
    to_check = endCNode
    while to_check != startCNode:
        for cnode in to_check._connections.keys():
            if cnode.cost + cnode._connections.get(to_check) == to_check.cost:
                path.append(cnode)
                to_check = cnode
                break
        else:
            raise CPathError("Coulnd't construct path")
        
    path.reverse()
    return path

# DJIKSTRA
@cython.boundscheck(False)
cpdef list djikstra_bestpath(CNode startCNode, CNode endCNode, bint first_contact = False):
    """
    Args:
        startCNode: starting point
        endCNode: ending point
        first_contact: if True, pathfinding will end as soon as the endCNode has first been discoverd
    
    Returns:
        list of CNodes, creating a path
    """
    cdef LowComby queue
    cdef CNode currentCNode
    cdef CNode cnode
    cdef double cost, new_cost
 
    # get costs
    startCNode.cost = 0
    queue = LowComby()
    queue.append(startCNode)
    
    while len(queue) != 0:
        currentCNode = wrappop(queue)
        for cnode, cost in currentCNode._connections.items():
            new_cost = currentCNode.cost + cost
            if new_cost < cnode.cost:
                cnode.cost = new_cost
                if cnode not in queue:
                    wrappush(queue, cnode)
        
        if endCNode.cost <= currentCNode.cost:
            break
        if first_contact and currentCNode is endCNode:
            break
    
    # get path
    return construct(startCNode, endCNode)
    
@cython.boundscheck(False)
cpdef list astar_bestpath(CNode startCNode, CNode endCNode, object func, bint first_contact = False, list args = []):
    """
    Args:
        startCNode: starting point
        endCNode: ending point
        first_contact: if True, pathfinding will end as soon as the endCNode has first been discoverd
    
    Returns:
        list of CNodes, creating a path
    """
    if args == []:
        args = []
    cdef LowComby queue
    cdef CNode currentCNode
    cdef CNode cnode
    cdef double cost, new_cost, result
 
    # get costs
    startCNode.cost = 0
    queue = LowComby()
    queue.append(startCNode)
    
    while len(queue) != 0:
        currentCNode = wrappop(queue)
        for cnode, cost in currentCNode._connections.items():
                
            new_cost = currentCNode.cost + cost
            if new_cost < cnode.cost:
                cnode.cost = new_cost
                result = func(cnode, args)
                cnode.probable_cost = new_cost + result
                if cnode not in queue:
                    wrappush(queue, cnode)
        
        if endCNode.cost <= currentCNode.cost:
            break
        if first_contact and currentCNode is endCNode:
            break
    
    # get path
    return construct(startCNode, endCNode)

cpdef tuple copy_graph(CNode start, CNode stop):
    """copies the Nodestructure deleting the found cost values for them"""
    cdef dict nodes = {}
    cdef CNode new_start = start._copy(nodes)
    return new_start, nodes.get(stop), nodes


from unittest import TestCase, main

from ex import example1, example2
from pypathfinder.Dijkstra import Node, bestpath
from pypathfinder.fast import CNode, djikstra_bestpath


class TestDijkstra(TestCase):
    def test_bestpath1(self):
        frankfurt, muenchen = example1(Node)
        path = bestpath(frankfurt, muenchen)
        self.assertEqual(path, [Node("Frankfurt"), Node("Würzburg"), Node("Nürnberg"), Node("München")])
        self.assertEqual(path[-1].cost, 487)
    
    def test_bestpath2(self):
        frankfurt, muenchen = example1(Node)
        path = bestpath(frankfurt, muenchen, True)
        self.assertEqual(path, [Node("Frankfurt"), Node("Würzburg"), Node("Nürnberg"), Node("München")])
        self.assertEqual(path[-1].cost, 487)
    
    def test_bestpath3(self):
        start, stop, matrix, solution = example2(Node, 2)
        path = bestpath(start, stop, True)
        self.assertEqual(path[-1].cost, solution)
    
    def test_cythonbestpath1(self):
        frankfurt, muenchen = example1(CNode)
        path = djikstra_bestpath(frankfurt, muenchen)
        self.assertEqual(path, [CNode("Frankfurt"), CNode("Würzburg"), CNode("Nürnberg"), CNode("München")])
        self.assertEqual(path[-1].cost, 487)
    
    def test_cythonbestpath2(self):
        frankfurt, muenchen = example1(CNode)
        path = djikstra_bestpath(frankfurt, muenchen, True)
        self.assertEqual(path, [CNode("Frankfurt"), CNode("Würzburg"), CNode("Nürnberg"), CNode("München")])
        self.assertEqual(path[-1].cost, 487)
    
    def test_cythonbestpath3(self):
        start, stop, matrix, solution = example2(CNode, 2)
        path = djikstra_bestpath(start, stop, True)
        self.assertEqual(path[-1].cost, solution)

if __name__ == "__main__":
    main()


from unittest import TestCase, main

from pypathfinder.Dijkstra import copy_graph, Node, bestpath
from pypathfinder import fast
from ex import example1

class TestDijkstra(TestCase):
    
    def test_copy1(self):
        point1 = Node(1)
        point2 = Node(2)
        point3 = Node(3)
        point4 = Node(4)

        point1.connect({point2: 10, point3: 5}, True)
        point4.connect({point2: 5, point3: 15}, True)

        path = bestpath(point1, point4) # Nodes will get a cost
        new_start, new_stop, all_nodes = copy_graph(point1, point4)
        self.assertEqual(point4.cost, 15)
        self.assertEqual(new_stop.cost, float("inf"))
        self.assertEqual(all_nodes.get(new_stop).cost, float("inf"))
        bestpath(new_start, new_stop)
        self.assertEqual(new_stop.cost, 15)

    def test_copy2(self):
        start, end = example1(Node)
        path = bestpath(start, end) # Nodes will get a cost
        new_start, new_end, all_nodes = copy_graph(start, end)
        self.assertEqual(end.cost, 487)
        self.assertEqual(new_end.cost, float("inf"))
    
    def test_cythoncopy1(self):
        start, end = example1(fast.CNode)
        path = bestpath(start, end) # Nodes will get a cost
        new_start, new_end, all_nodes = fast.copy_graph(start, end)
        self.assertEqual(end.cost, 487)
        self.assertEqual(new_end.cost, float("inf"))


if __name__ == "__main__":
    main()


from setuptools import setup, Extension

CPATH = "pypathfinder\\fast"
CPACK = CPATH.replace("\\", ".")

ext1 = Extension(f"{CPACK}.ctools", [f"{CPATH}\\ctools.c"])
ext = [ext1]

setup(
    ext_modules=ext
)



from typing import Dict, Hashable, List, Tuple, Union
from pypathfinder.utils import HighComby, LowComby, PathError, PathfinderError, get_pop, get_push
from heapq import heappop, heappush
from functools import total_ordering

@total_ordering
class Node:
    __slots__ = ("_connections", "id", "cost")
    def __init__(self, id: Hashable, connections: Dict["Node", int] = None):
        self._connections: Dict[Node, int] = {} if connections is None else connections
        self.id = id
        self.cost = float("inf")
    
    def __str__(self) -> str:
        return f"Node(id={self.id}, cost={self.cost})"
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def __lt__(self, other):
        if not isinstance(other, type(self)):
            raise NotImplemented
        return self.cost < other.cost
    
    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, type(self)):
            return False
        return __o.id == self.id

    def __hash__(self) -> int:
        return hash(self.id)
    
    def connect(self, conn: Dict["Node", int], reflect: bool = False):
        """
        Args:
            conn: connections
            reflect: if the connections should also be applied to the connected Nodes
        """
        self._connections.update(conn)
        if reflect:
            for node, value in conn.items():
                node.connect({self: value}, False)

    def _copy(self, nodes: dict) -> "Node": 
        self_copy = type(self)(self.id)
        nodes[self_copy] = self_copy
        new_connections = {}
        for node, cost in self._connections.items():
            if node in nodes.keys():
                node_copy = nodes.get(node)
            else:
                node_copy = node._copy(nodes)
                nodes[node_copy] = node_copy
            new_connections[node_copy] = cost
        self_copy.connect(new_connections, False)
        return self_copy
        #return self_copy

def construct(startnode, endnode) -> list:
    # get path
    path = [endnode]
    to_check = endnode
    while to_check != startnode:
        for node in to_check._connections.keys():
            if node.cost + node._connections.get(to_check) == to_check.cost:
                path.append(node)
                to_check = node
                break
        else:
            raise PathError("Coulnd't construct path")
        
    path.reverse()
    return path

def bestpath(startnode: Node, endnode: Node, first_contact: bool = False, queue_type: Union[list, LowComby, HighComby]=list) -> List[Node]:
    """
    Args:
        startnode: starting point
        endnode: ending point
        first_contact: if True, pathfinding will end as soon as the endnode has first been discoverd
        queue_type: type used for queueing nodes; LowComby can be faster but consumes more memory
    
    Returns:
        list of nodes, creating a path
    """
    # get costs
    startnode.cost = 0
    queue: Union[List[Node], LowComby] = queue_type()
    queue.append(startnode)
    use_heappop = get_pop(queue)
    use_heappush = get_push(queue)
    
    while len(queue) != 0:
        currentnode = use_heappop(queue)
        for node, cost in currentnode._connections.items():
                
            new_cost = currentnode.cost + cost
            if new_cost < node.cost:
                node.cost = new_cost
                if node not in queue:
                    use_heappush(queue, node)
        
        if endnode.cost <= currentnode.cost:
            break
        if first_contact and currentnode is endnode:
            break
    
    # get path
    return construct(startnode, endnode)

def copy_graph(start: Node, stop: Node) -> Tuple[Node, Node, Dict[str, Node]]:
    """copies the Nodestructure deleting the found cost values for them"""
    nodes = {}
    new_start = start._copy(nodes)
    return new_start, nodes.get(stop), nodes




from unittest import TestCase, main

from ex import example2 # needed example
from pypathfinder.Astar import ANode, bestpath
from pypathfinder.fast import CNode, astar_bestpath

class TestAstar(TestCase):
    def test_bestpath1(self):
        start, stop, matrix, solution = example2(ANode, 0)
        def heusterik(node: ANode, args) -> int:
            return abs(node.id[0] - len(matrix[0])) + abs(node.id[1] - len(matrix))
        path = bestpath(start, stop, heusterik)
        self.assertEqual(path[-1].cost, solution)
    
    def test_bestpath2(self):
        start, stop, matrix, solution = example2(ANode, 2)
        def heusterik(node: ANode, args) -> int:
            return abs(node.id[0] - len(matrix[0])) + abs(node.id[1] - len(matrix))
        path = bestpath(start, stop, heusterik)
        self.assertEqual(path[-1].cost, solution)

    def test_cythonbestpath1(self):
        start, stop, matrix, solution = example2(CNode, 0)
        def heusterik(node: CNode, args) -> int:
            return abs(node.id[0] - len(matrix[0])) + abs(node.id[1] - len(matrix))
        path = astar_bestpath(start, stop, heusterik)
        self.assertEqual(path[-1].cost, solution)
    
    def test_cythonbestpath2(self):
        start, stop, matrix, solution = example2(CNode, 2)
        def heusterik(node: CNode, args) -> int:
            return abs(node.id[0] - len(matrix[0])) + abs(node.id[1] - len(matrix))
        path = astar_bestpath(start, stop, heusterik)
        self.assertEqual(path[-1].cost, solution)
    
if __name__ == "__main__":
    main()